"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import abc
import uuid
from typing import TYPE_CHECKING, Dict, Optional, Union

from ..asset import Asset
from ..enums import Locale
from ..localization import Localization

if TYPE_CHECKING:
    from ..cache import CacheState
    from ..types.object import GridPosition as GridPositionPayload, ShopData as ShopDataPayload
    from .gear import Gear
    from .weapons import Weapon


__all__ = (
    'BaseModel',
    'GridPosition',
    'ShopData',
)


class BaseModel(abc.ABC):
    __slots__ = ('_uuid',)

    def __init__(self, uuid: str) -> None:
        self._uuid = uuid

    @property
    def uuid(self) -> uuid.UUID:
        return uuid.UUID(self._uuid)

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} uuid={self.uuid!r}>'

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.uuid == other.uuid

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.uuid)


class GridPosition:
    def __init__(self, data: GridPositionPayload) -> None:
        self.row: float = data['row']
        self.column: float = data['column']

    def __repr__(self) -> str:
        return f'<GridPosition row={self.row} column={self.column}>'

    def __eq__(self, other: object) -> bool:
        return isinstance(other, GridPosition) and self.row == other.row and self.column == other.column

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)


class ShopData:
    def __init__(self, *, state: CacheState, item: Union[Weapon, Gear], data: ShopDataPayload) -> None:
        self._item: Union[Weapon, Gear] = item
        self._state: CacheState = state
        self.cost: int = data['cost']
        self.category: Optional[str] = data['category']
        self._category_text: Union[str, Dict[str, str]] = data['categoryText']
        self.grid_position: Optional[GridPosition] = None
        if data['gridPosition'] is not None:
            self.grid_position = GridPosition(data['gridPosition'])
        self.can_be_trashed: bool = data['canBeTrashed']
        self._image: Optional[str] = data['image']
        self._new_image: Optional[str] = data['newImage']
        self._new_image_2: Optional[str] = data['newImage2']
        self.asset_path: str = data['assetPath']
        self._category_text_localized: Localization = Localization(self._category_text, locale=self._state.locale)

    def __repr__(self) -> str:
        return f'<ShopData category_text={self.category_text} cost={self.cost}>'

    def __int__(self) -> int:
        return self.cost

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ShopData) and self.item == other.item and self.cost == other.cost

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def category_text_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._category_text_localized.from_locale(locale)

    @property
    def item(self) -> Union[Weapon, Gear]:
        """:class: `Weapon` or :class: `Gear` Returns the item."""
        return self._item

    @property
    def category_text(self) -> Localization:
        """:class: `str` Returns the weapon's shop category text."""
        return self._category_text_localized

    @property
    def image(self) -> Optional[Asset]:
        """:class: `Asset` Returns the weapon's image."""
        return Asset._from_url(self._state, url=self._image) if self._image else None

    @property
    def new_image(self) -> Optional[Asset]:
        """:class: `Asset` Returns the weapon's new image."""
        return Asset._from_url(self._state, url=self._new_image) if self._new_image else None

    @property
    def new_image_2(self) -> Optional[Asset]:
        """:class: `Asset` Returns the weapon's new image 2."""
        return Asset._from_url(self._state, url=self._new_image_2) if self._new_image_2 else None



"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from typing import Dict, List, Optional, TypedDict, Union

from .object import Object, ShopData
from .response import Response


class AdsStats(TypedDict):
    zoomMultiplier: float
    fireRate: float
    runSpeedMultiplier: float
    burstCount: int
    firstBulletAccuracy: float


class AltShotgunStats(TypedDict):
    shotgunPelletCount: int
    burstRate: float


class AirBurstStats(TypedDict):
    shotgunPelletCount: int
    burstDistance: float


class DamageRange(TypedDict):
    rangeStartMeters: float
    rangeEndMeters: float
    headDamage: float
    bodyDamage: float
    legDamage: float


class WeaponStats(TypedDict):
    fireRate: float
    magazineSize: int
    runSpeedMultiplier: float
    equipTimeSeconds: float
    reloadTimeSeconds: float
    firstBulletAccuracy: float
    shotgunPelletCount: int
    wallPenetration: str
    feature: Optional[str]
    fireMode: Optional[str]
    altFireType: Optional[str]
    adsStats: Optional[AdsStats]
    altShotgunStats: Optional[AltShotgunStats]
    airBurstStats: Optional[AirBurstStats]
    damageRanges: List[DamageRange]


class Chroma(Object):
    displayName: Union[str, Dict[str, str]]
    displayIcon: Optional[str]
    fullRender: str
    swatch: Optional[str]
    streamedVideo: Optional[str]
    assetPath: str


class Level(Object):
    displayName: Union[str, Dict[str, str]]
    levelItem: str
    displayIcon: Optional[str]
    streamedVideo: Optional[str]
    assetPath: str


class Skin(Object):
    displayName: Union[str, Dict[str, str]]
    themeUuid: str
    contentTierUuid: str
    displayIcon: str
    wallpaper: str
    assetPath: str
    chromas: List[Chroma]
    levels: List[Level]


class Weapon(Object):
    displayName: Union[str, Dict[str, str]]
    category: str
    defaultSkinUuid: str
    displayIcon: str
    killStreamIcon: str
    assetPath: str
    weaponStats: Optional[WeaponStats]
    shopData: Optional[ShopData]
    skins: List[Skin]


Weapons = Response[List[Weapon]]
WeaponUUID = Response[Weapon]

Skins = Response[List[Skin]]
SkinUUID = Response[Skin]

SkinLevels = Response[List[Level]]
SkinLevel = Level
SkinLevelUUID = Response[Level]

SkinChromas = Response[List[Chroma]]
SkinChroma = Chroma
SkinChromaUUID = Response[Chroma]



"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional, Union

from .. import utils
from ..asset import Asset
from ..enums import Locale
from ..localization import Localization
from .abc import BaseModel

if TYPE_CHECKING:
    from typing_extensions import Self

    from ..cache import CacheState
    from ..types.sprays import Spray as SprayPayload, SprayLevel as SprayLevelPayload
    from .themes import Theme

__all__ = (
    'Spray',
    'SprayLevel',
)


class Spray(BaseModel):
    def __init__(self, *, state: CacheState, data: SprayPayload) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self._data: SprayPayload = data
        self._display_name: Union[str, Dict[str, str]] = data['displayName']
        self._category: Optional[str] = data['category']
        self._theme_uuid: Optional[str] = data['themeUuid']
        self._is_null_spray: bool = data['isNullSpray']
        self._display_icon: str = data['displayIcon']
        self._full_icon: Optional[str] = data['fullIcon']
        self._full_transparent_icon: Optional[str] = data['fullTransparentIcon']
        self._animation_png: Optional[str] = data['animationPng']
        self._animation_gif: Optional[str] = data['animationGif']
        self.asset_path: str = data['assetPath']
        self.levels: List[SprayLevel] = [SprayLevel(state=self._state, data=level, parent=self) for level in data['levels']]
        self._display_name_localized: Localization = Localization(self._display_name, locale=self._state.locale)

    def __str__(self) -> str:
        return self.display_name.locale

    def __repr__(self) -> str:
        return f'<Spray display_name={self.display_name!r}>'

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._display_name_localized.from_locale(locale)

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the skin's name."""
        return self._display_name_localized

    @property
    def category(self) -> Optional[str]:
        """:class: `str` Returns the skin's category."""
        if self._category is None:
            return None
        return utils.removeprefix(self._category, 'EAresSprayCategory::')

    @property
    def theme(self) -> Optional[Theme]:
        if self._theme_uuid is None:
            return None
        return self._state.get_theme(self._theme_uuid)

    @property
    def display_icon(self) -> Optional[Asset]:
        """:class: `Asset` Returns the skin's icon."""
        return Asset._from_url(state=self._state, url=self._display_icon)

    @property
    def full_icon(self) -> Optional[Asset]:
        """:class: `Asset` Returns the skin's full icon."""
        if self._full_icon is None:
            return None
        return Asset._from_url(state=self._state, url=self._full_icon)

    @property
    def full_transparent_icon(self) -> Optional[Asset]:
        """:class: `Asset` Returns the skin's full transparent icon."""
        if self._full_transparent_icon is None:
            return None
        return Asset._from_url(state=self._state, url=self._full_transparent_icon)

    @property
    def animation_png(self) -> Optional[Asset]:
        """:class: `Asset` Returns the skin's animation png."""
        if self._animation_png is None:
            return None
        return Asset._from_url(state=self._state, url=self._animation_png)

    @property
    def animation_gif(self) -> Optional[Asset]:
        """:class: `Asset` Returns the skin's animation gif."""
        if self._animation_gif is None:
            return None
        return Asset._from_url(state=self._state, url=self._animation_gif)

    def is_null_spray(self) -> bool:
        """:class: `bool` Returns a boolean representing whether the skin is a null spray."""
        return self._is_null_spray

    @classmethod
    def _copy(cls, spray: Self) -> Self:
        self = cls.__new__(cls)  # bypass __init__
        self._uuid = spray._uuid
        self._state = spray._state
        self._data = spray._data.copy()
        self._display_name = spray._display_name
        self._category = spray._category
        self._theme_uuid = spray._theme_uuid
        self._is_null_spray = spray._is_null_spray
        self._display_icon = spray._display_icon
        self._full_icon = spray._full_icon
        self._full_transparent_icon = spray._full_transparent_icon
        self._animation_png = spray._animation_png
        self._animation_gif = spray._animation_gif
        self.asset_path = spray.asset_path
        self.levels = spray.levels.copy()
        self._display_name_localized = spray._display_name_localized
        return self


class SprayLevel(BaseModel):
    def __init__(self, state: CacheState, data: SprayLevelPayload, parent: Spray) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self._data: SprayLevelPayload = data
        self.spray_level: int = data['sprayLevel']
        self._display_name: Union[str, Dict[str, str]] = data['displayName']
        self._display_icon: Optional[str] = data['displayIcon']
        self.asset_path: str = data['assetPath']
        self.parent: Spray = parent
        self._display_name_localized: Localization = Localization(self._display_name, locale=self._state.locale)

    def __str__(self) -> str:
        return self.display_name.locale

    def __repr__(self) -> str:
        return f'<SprayLevel display_name={self.display_name!r}>'

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._display_name_localized.from_locale(locale)

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the spray's name."""
        return self._display_name_localized

    @property
    def level(self) -> int:
        """:class: `int` alias for :attr: `SprayLevel.spray_level`"""
        return self.spray_level

    @property
    def display_icon(self) -> Optional[Asset]:
        """:class: `str` Returns the spray's display icon."""
        return Asset._from_url(state=self._state, url=self._display_icon) if self._display_icon else None

    @classmethod
    def _copy(cls, spray_level: Self) -> Self:
        self = cls.__new__(cls)  # bypass __init__
        self._uuid = spray_level._uuid
        self._state = spray_level._state
        self._data = spray_level._data.copy()
        self.spray_level = spray_level.spray_level
        self._display_name = spray_level._display_name
        self._display_icon = spray_level._display_icon
        self.asset_path = spray_level.asset_path
        self.parent = spray_level.parent._copy(spray_level.parent)
        self._display_name_localized = spray_level._display_name_localized
        return self



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz
Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import types
from collections import namedtuple
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Iterator, List, Mapping, Optional, Tuple, Type, TypeVar

# enums class for discord.py: https://github.com/Rapptz/discord.py/blob/master/discord/asset.py


def _create_value_cls(name: str, comparable: bool):
    # All the type ignores here are due to the type checker being unable to recognise
    # Runtime type creation without exploding.
    cls = namedtuple('_EnumValue_' + name, 'name value')
    cls.__repr__ = lambda self: f'<{name}.{self.name}: {self.value!r}>'  # type: ignore
    cls.__str__ = lambda self: f'{name}.{self.name}'  # type: ignore
    if comparable:
        cls.__le__ = lambda self, other: isinstance(other, self.__class__) and self.value <= other.value  # type: ignore
        cls.__ge__ = lambda self, other: isinstance(other, self.__class__) and self.value >= other.value  # type: ignore
        cls.__lt__ = lambda self, other: isinstance(other, self.__class__) and self.value < other.value  # type: ignore
        cls.__gt__ = lambda self, other: isinstance(other, self.__class__) and self.value > other.value  # type: ignore
    return cls


def _is_descriptor(obj):
    return hasattr(obj, '__get__') or hasattr(obj, '__set__') or hasattr(obj, '__delete__')


class EnumMeta(type):
    if TYPE_CHECKING:
        __name__: ClassVar[str]
        _enum_member_names_: ClassVar[List[str]]
        _enum_member_map_: ClassVar[Dict[str, Any]]
        _enum_value_map_: ClassVar[Dict[Any, Any]]
        _enum_string_key_map_: ClassVar[Dict[str, Any]]

    def __new__(cls, name: str, bases: Tuple[type, ...], attrs: Dict[str, Any], *, comparable: bool = False) -> Self:
        value_mapping = {}
        string_key_mapping = {}
        member_mapping = {}
        member_names = []

        value_cls = _create_value_cls(name, comparable)
        for key, value in list(attrs.items()):
            is_descriptor = _is_descriptor(value)
            if key[0] == '_' and not is_descriptor:
                continue

            # Special case classmethod to just pass through
            if isinstance(value, classmethod):
                continue

            if is_descriptor:
                setattr(value_cls, key, value)
                del attrs[key]
                continue

            try:
                new_value = value_mapping[value]
            except KeyError:
                new_value = value_cls(name=key, value=value)
                value_mapping[value] = new_value
                member_names.append(key)

            try:
                string_key_mapping[value]
            except KeyError:
                string_key_mapping[str(key)] = value_cls(name=key, value=value)

            member_mapping[key] = new_value
            attrs[key] = new_value

        attrs['_enum_value_map_'] = value_mapping
        attrs['_enum_string_key_map_'] = string_key_mapping
        attrs['_enum_member_map_'] = member_mapping
        attrs['_enum_member_names_'] = member_names
        attrs['_enum_value_cls_'] = value_cls
        actual_cls = super().__new__(cls, name, bases, attrs)
        value_cls._actual_enum_cls_ = actual_cls  # type: ignore # Runtime attribute isn't understood
        return actual_cls

    def __iter__(cls) -> Iterator[Any]:
        return (cls._enum_member_map_[name] for name in cls._enum_member_names_)

    def __reversed__(cls) -> Iterator[Any]:
        return (cls._enum_member_map_[name] for name in reversed(cls._enum_member_names_))

    def __len__(cls) -> int:
        return len(cls._enum_member_names_)

    def __repr__(cls) -> str:
        return f'<enum {cls.__name__}>'

    @property
    def __members__(cls) -> Mapping[str, Any]:
        return types.MappingProxyType(cls._enum_member_map_)

    def __call__(cls, value: str) -> Any:
        try:
            return cls._enum_value_map_[value]
        except (KeyError, TypeError):
            raise ValueError(f"{value!r} is not a valid {cls.__name__}")

    def __getitem__(cls, key: str) -> Any:
        return cls._enum_member_map_[key]

    def __setattr__(cls, name: str, value: Any) -> None:
        raise TypeError('Enums are immutable.')

    def __delattr__(cls, attr: str) -> None:
        raise TypeError('Enums are immutable')

    def __instancecheck__(self, instance: Any) -> bool:
        # isinstance(x, Y)
        # -> __instancecheck__(Y, x)
        try:
            return instance._actual_enum_cls_ is self
        except AttributeError:
            return False


if TYPE_CHECKING:
    from enum import Enum

    from typing_extensions import Self
else:

    class Enum(metaclass=EnumMeta):
        @classmethod
        def try_value(cls, value):
            try:
                return cls._enum_value_map_[value]
            except (KeyError, TypeError):
                return value


# --

__all__ = (
    'MELEE_WEAPON_ID',
    'AbilityType',
    'Locale',
    'MissionType',
    'RelationType',
    'RewardType',
    'try_enum',
)

MELEE_WEAPON_ID: str = '2f59173c-4bed-b6c3-2191-dea9b58be9c7'


class AbilityType(Enum):
    passive = 'Passive'
    grenade = 'Grenade'
    ability_1 = 'Ability1'
    ability_2 = 'Ability2'
    ultimate = 'Ultimate'


class RelationType(Enum):
    agent = 'Agent'
    event = 'Event'
    season = 'Season'
    none = 'None'

    def __str__(self) -> str:
        return str(self.value)


class RewardType(Enum):
    skin_level = 'EquippableSkinLevel'
    buddy_level = 'EquippableCharmLevel'
    currency = 'Currency'
    player_card = 'PlayerCard'
    player_title = 'Title'
    spray = 'Spray'
    agent = 'Character'

    def __str__(self) -> str:
        return str(self.value)


class MissionType(Enum):
    weekly = 'Weekly'
    daily = 'Daily'
    tutorial = 'Tutorial'
    npe = 'NPE'

    def __str__(self) -> str:
        return str(self.value)

    @property
    def full(self) -> str:
        return f'AresMissionType::{self.value}'


class Locale(Enum):
    arabic = 'ar-AE'
    german = 'de-DE'
    american_english = 'en-US'
    british_english = 'en-US'
    spain_spanish = 'es-ES'
    spanish_mexican = 'es-MX'
    french = 'fr-FR'
    indonesian = 'id-ID'
    italian = 'it-IT'
    japanese = 'ja-JP'
    korean = 'ko-KR'
    polish = 'pl-PL'
    brazil_portuguese = 'pt-BR'
    russian = 'ru-RU'
    thai = 'th-TH'
    turkish = 'tr-TR'
    vietnamese = 'vi-VN'
    chinese = 'zh-CN'
    taiwan_chinese = 'zh-TW'

    # aliases
    english = 'en-US'

    def __str__(self) -> str:
        return str(self.value)


# from discord.py
# https://github.com/Rapptz/discord.py/blob/master/discord/enums.py

E = TypeVar('E', bound='Enum')


def create_unknown_value(cls: Type[E], val: Any) -> E:
    value_cls = cls._enum_value_cls_  # type: ignore # This is narrowed below
    name = f'unknown_{val}'
    return value_cls(name=name, value=val)


def try_enum(cls: Type[E], val: Any, default: Optional[Any] = None) -> E:
    """A function that tries to turn the value into enum ``cls``.

    If it fails it returns a proxy invalid value instead.
    """
    try:
        return cls._enum_value_map_[val]  # type: ignore # All errors are caught below
    except (KeyError, TypeError, AttributeError):
        if default is not None:
            return default
        return create_unknown_value(cls, val)


# ---



"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional, Union

if TYPE_CHECKING:
    from aiohttp import ClientResponse

__all__ = (
    'BadRequest',
    'Forbidden',
    'HTTPException',
    'InternalServerError',
    'NotFound',
    'RateLimited',
    'ValorantError',
)


class ValorantError(Exception):
    """Exception that's raised when a Valorant API request operation fails."""

    pass


class HTTPException(ValorantError):
    """Exception that's raised when an HTTP request operation fails.
    Attributes
    ------------
    response: :class:`aiohttp.ClientResponse`
        The response of the failed HTTP request. This is an
        instance of :class:`aiohttp.ClientResponse`. In some cases
        this could also be a :class:`requests.Response`.
    text: :class:`str`
        The text of the error. Could be an empty string.
    status: :class:`int`
        The status code of the HTTP request.
    """

    def __init__(self, response: ClientResponse, message: Optional[Union[str, Dict[str, Any]]]):
        self.response: ClientResponse = response
        self.status: int = response.status  # This attribute is filled by the library even if using requests # noqa: E501
        self.text: str
        if isinstance(message, dict):
            self.text = message.get('message') or message.get('error', '')
        else:
            self.text = message or ''

        fmt = '{0.status} {0.reason}'
        if len(self.text):
            fmt += ': {1}'

        super().__init__(fmt.format(self.response, self.text))


class BadRequest(HTTPException):
    """Exception that's raised for when status code 400 occurs.
    Subclass of :exc:`HTTPException`
    """

    pass


class Forbidden(HTTPException):
    """Exception that's raised for when status code 403 occurs.
    Subclass of :exc:`HTTPException`
    """

    pass


class NotFound(HTTPException):
    """Exception that's raised for when status code 404 occurs.
    Subclass of :exc:`HTTPException`
    """

    pass


class InternalServerError(HTTPException):
    """Exception that's raised for when status code 500 occurs.
    Subclass of :exc:`HTTPException`
    """

    pass


class RateLimited(HTTPException):
    """Exception that's raised for when a 429 status code occurs.
    Subclass of :exc:`HTTPException`.
    """

    pass



from __future__ import annotations

"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from typing import TYPE_CHECKING

from ..asset import Asset
from .abc import BaseModel

if TYPE_CHECKING:
    from typing_extensions import Self

    from ..cache import CacheState
    from ..types.level_borders import LevelBorder as LevelBorderPayload

# fmt: off
__all__ = (
    'LevelBorder',
)
# fmt: on


class LevelBorder(BaseModel):
    def __init__(self, state: CacheState, data: LevelBorderPayload) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self.starting_level: int = data['startingLevel']
        self._level_number_appearance: str = data['levelNumberAppearance']
        self._small_player_card_appearance: str = data['smallPlayerCardAppearance']
        self.asset_path: str = data['assetPath']

    def __repr__(self) -> str:
        return f'<LevelBorder starting_level={self.starting_level!r}>'

    def __lt__(self, other: object) -> bool:
        return isinstance(other, LevelBorder) and self.starting_level < other.starting_level

    def __le__(self, other: object) -> bool:
        return isinstance(other, LevelBorder) and self.starting_level <= other.starting_level

    def __gt__(self, other: object) -> bool:
        return isinstance(other, LevelBorder) and self.starting_level > other.starting_level

    def __ge__(self, other: object) -> bool:
        return isinstance(other, LevelBorder) and self.starting_level >= other.starting_level

    @property
    def level_number_appearance(self) -> Asset:
        """:class: `Asset` Returns the level number appearance of the level border."""
        return Asset._from_url(state=self._state, url=self._level_number_appearance)

    @property
    def small_player_card_appearance(self) -> Asset:
        """:class: `Asset` Returns the small player card appearance of the level border."""
        return Asset._from_url(state=self._state, url=self._small_player_card_appearance)

    @classmethod
    def _copy(cls, level_border: Self) -> Self:
        """Returns a copy of the level border."""
        self = cls.__new__(cls)  # bypass __init__
        self._uuid = level_border._uuid
        self._state = level_border._state
        self.starting_level = level_border.starting_level
        self._level_number_appearance = level_border._level_number_appearance
        self._small_player_card_appearance = level_border._small_player_card_appearance
        self.asset_path = level_border.asset_path
        return self



"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional, Union

from ..asset import Asset
from ..enums import Locale
from ..localization import Localization
from .abc import BaseModel

if TYPE_CHECKING:
    from ..cache import CacheState
    from ..types.bundles import Bundle as BundlePayload
    from .buddies import Buddy
    from .player_cards import PlayerCard
    from .player_titles import PlayerTitle
    from .sprays import Spray
    from .weapons import Skin

    BundleItem = Union[Skin, Buddy, Spray, PlayerCard, PlayerTitle]

# fmt: off
__all__ = (
    'Bundle',
)
# fmt: on


class Bundle(BaseModel):
    def __init__(self, *, state: CacheState, data: BundlePayload) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self._data: BundlePayload = data
        self._display_name: Union[str, Dict[str, str]] = data['displayName']
        self._display_name_sub_text: Optional[Union[str, Dict[str, str]]] = data['displayNameSubText']
        self._description: Union[str, Dict[str, str]] = data['description']
        self._extra_description: Optional[Union[str, Dict[str, str]]] = data['extraDescription']
        self._promo_description: Optional[Union[str, Dict[str, str]]] = data['promoDescription']
        self.use_additional_context: bool = data['useAdditionalContext']
        self._display_icon: str = data['displayIcon']
        self._display_icon_2: str = data['displayIcon2']
        self._vertical_promo_image: str = data['verticalPromoImage']
        self.asset_path: str = data['assetPath']
        self._display_name_localized: Localization = Localization(self._display_name, locale=self._state.locale)
        self._display_name_sub_text_localized: Localization = Localization(
            self._display_name_sub_text, locale=self._state.locale
        )
        self._description_localized: Localization = Localization(self._description, locale=self._state.locale)
        self._extra_description_localized: Localization = Localization(self._extra_description, locale=self._state.locale)
        self._promo_description_localized: Localization = Localization(self._promo_description, locale=self._state.locale)
        self._items: List[BundleItem] = []

    def __str__(self) -> str:
        return self.display_name.locale

    def __repr__(self) -> str:
        return f'<Bundle display_name={self.display_name!r}>'

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._display_name_localized.from_locale(locale)

    def display_name_sub_text_localized(self, locale: Optional[Union[Locale, str]] = None) -> Optional[str]:
        if self._description is None:
            return None
        return self._display_name_sub_text_localized.from_locale(locale)

    def description_localized(self, locale: Optional[Union[Locale, str]] = None) -> Optional[str]:
        if self._description is None:
            return None
        return self._description_localized.from_locale(locale)

    def extra_description_localized(self, locale: Optional[Union[Locale, str]] = None) -> Optional[str]:
        if self._extra_description is None:
            return None
        return self._extra_description_localized.from_locale(locale)

    def promo_description_localized(self, locale: Optional[Union[Locale, str]] = None) -> Optional[str]:
        if self._promo_description is None:
            return None
        return self._promo_description_localized.from_locale(locale)

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the bundle's name."""
        return self._display_name_localized

    @property
    def display_name_sub_text(self) -> Optional[Localization]:
        """:class: `str` Returns the bundle's sub name."""
        if self._display_name_sub_text is None:
            return None
        return self._display_name_sub_text_localized

    @property
    def description(self) -> Optional[Localization]:
        """:class: `str` Returns the bundle's description."""
        if self._description is None:
            return None
        return self._description_localized

    @property
    def extra_description(self) -> Optional[Localization]:
        """:class: `str` Returns the bundle's description extra localizations."""
        if self._extra_description is None:
            return None
        return self._extra_description_localized

    @property
    def promo_description(self) -> Optional[Localization]:
        """:class: `str` Returns the bundle's description promo."""
        if self._promo_description is None:
            return None
        return self._promo_description_localized

    @property
    def display_icon(self) -> Asset:
        """:class: `Asset` Returns the bundle's icon."""
        return Asset._from_url(state=self._state, url=self._display_icon)

    @property
    def display_icon_2(self) -> Asset:
        """:class: `Asset` Returns the bundle's icon 2."""
        return Asset._from_url(state=self._state, url=self._display_icon_2)

    @property
    def vertical_promo_image(self) -> Optional[Asset]:
        """:class: `Asset` Returns the bundle's vertical promo image."""
        if self._vertical_promo_image is None:
            return None
        return Asset._from_url(state=self._state, url=self._vertical_promo_image)

    def _add_item(self, item: Union[Skin, Buddy, Spray, PlayerCard]) -> None:
        self._items.append(item)

    @property
    def items(self) -> List[BundleItem]:
        """:class: `List[Union[Buddy, Spray, PlayerCard]]` Returns the bundle's items."""
        return self._items

    @property
    def buddies(self) -> List[Buddy]:
        """:class: `List[Buddy]` Returns the bundle's buddies."""
        # avoid circular imports
        from .buddies import Buddy

        return [item for item in self._items if isinstance(item, Buddy)]

    @property
    def sprays(self) -> List[Spray]:
        """:class: `List[Spray]` Returns the bundle's sprays."""
        # avoid circular imports
        from .sprays import Spray

        return [item for item in self._items if isinstance(item, Spray)]

    @property
    def player_cards(self) -> List[PlayerCard]:
        """:class: `List[PlayerCard]` Returns the bundle's player cards."""
        # avoid circular imports
        from .player_cards import PlayerCard

        return [item for item in self._items if isinstance(item, PlayerCard)]

    @property
    def skins(self) -> List[Skin]:
        """:class: `List[Skin]` Returns the bundle's skins."""
        # avoid circular imports
        from .weapons import Skin

        return [item for item in self._items if isinstance(item, Skin)]



"""
valorant.models
~~~~~~~~~~~~~~

Typings for the Valorant API

:copyright: (c) 2023-present STACiA
:license: MIT, see LICENSE for more details.

"""
from . import abc as abc
from .agents import *
from .buddies import *
from .bundles import *
from .ceremonies import *
from .competitive_tiers import *
from .content_tiers import *
from .contracts import *
from .currencies import *
from .events import *
from .gamemodes import *
from .gear import *
from .level_borders import *
from .maps import *
from .missions import *
from .player_cards import *
from .player_titles import *
from .seasons import *
from .sprays import *
from .themes import *
from .version import *
from .weapons import *



"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from ..asset import Asset
from ..enums import Locale, RelationType, RewardType, try_enum
from ..localization import Localization
from .abc import BaseModel

if TYPE_CHECKING:
    from uuid import UUID

    from ..cache import CacheState
    from ..types.contracts import (
        Chapter as ChapterPayload,
        Content as ContentPayload,
        Contract as ContractPayload,
        Level as LevelPayload,
        Reward as RewardPayload,
    )
    from .agents import Agent
    from .buddies import BuddyLevel
    from .currencies import Currency
    from .events import Event
    from .player_cards import PlayerCard
    from .player_titles import PlayerTitle
    from .seasons import Season
    from .sprays import Spray
    from .weapons import SkinLevel

# fmt: off
__all__ = (
    'Contract',
)
# fmt: on

_log = logging.getLogger(__name__)


class Reward(BaseModel):
    def __init__(self, state: CacheState, data: RewardPayload, chapter: Chapter, free_reward: bool = False) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self.type: RewardType = try_enum(RewardType, data['type'])
        self.amount: int = data['amount']
        self._is_highlighted: bool = data['isHighlighted']
        self._is_free: bool = free_reward
        self.chapter: Chapter = chapter

    def get_item(self) -> Optional[Union[Agent, SkinLevel, BuddyLevel, Currency, PlayerCard, PlayerTitle, Spray]]:
        if self.type is RewardType.skin_level:
            return self._state.get_skin_level(str(self.uuid))
        elif self.type is RewardType.buddy_level:
            return self._state.get_buddy_level(str(self.uuid))
        elif self.type is RewardType.player_card:
            return self._state.get_player_card(str(self.uuid))
        elif self.type is RewardType.player_title:
            return self._state.get_player_title(str(self.uuid))
        elif self.type is RewardType.spray:
            return self._state.get_spray(str(self.uuid))
        elif self.type is RewardType.agent:
            return self._state.get_agent(str(self.uuid))
        elif self.type is RewardType.currency:
            return self._state.get_currency(str(self.uuid))
        _log.warning(f'Unknown reward type: {self.type}')
        return None

    def is_highlighted(self) -> bool:
        return self._is_highlighted

    def is_free(self) -> bool:
        return self._is_free


class Level:
    def __init__(self, state: CacheState, data: LevelPayload, chapter: Chapter) -> None:
        self._state: CacheState = state
        self.reward: Reward = Reward(self._state, data['reward'], chapter)
        self.xp: int = data['xp']
        self.vp_cost: int = data['vpCost']
        self._is_purchasable_with_vp: bool = data['isPurchasableWithVP']
        self.dough_cost: int = data['doughCost']
        self._is_purchasable_with_dough: bool = data['isPurchasableWithDough']
        self.chapter: Chapter = chapter

    def is_purchasable_with_vp(self) -> bool:
        return self._is_purchasable_with_vp

    def is_purchasable_with_dough(self) -> bool:
        return self._is_purchasable_with_dough


class Chapter:
    def __init__(self, state: CacheState, data: ChapterPayload, index: int) -> None:
        self._state: CacheState = state
        self._is_epilogue: bool = data['isEpilogue']
        self.levels: List[Level] = [Level(self._state, level, self) for level in data['levels']]
        self.free_rewards: Optional[List[Reward]] = None
        if data['freeRewards'] is not None:
            self.free_rewards = [Reward(self._state, reward, self, free_reward=True) for reward in data['freeRewards']]
        self.index: int = index

    def is_epilogue(self) -> bool:
        return self._is_epilogue

    # helper methods

    def get_all_rewards(self) -> List[Reward]:
        """:class: `List[Reward]` Returns all rewards."""
        return [level.reward for level in self.levels]


class Content:
    def __init__(self, state: CacheState, data: ContentPayload) -> None:
        self._state: CacheState = state
        if data['relationType'] is None:
            self.relation_type: RelationType = RelationType.none
        else:
            self.relation_type = try_enum(RelationType, data['relationType'])
        self._relation_uuid: Optional[str] = data['relationUuid']
        self._chapters: List[Chapter] = [
            Chapter(self._state, chapter, index) for index, chapter in enumerate(data['chapters'])
        ]
        self.premium_reward_schedule_uuid: Optional[str] = data['premiumRewardScheduleUuid']
        self.premium_vp_cost: int = data['premiumVPCost']

    @property
    def chapters(self) -> List[Chapter]:
        return self._chapters

    @property
    def relationship(self) -> Optional[Union[Agent, Event, Season]]:
        if self.relation_type is RelationType.agent:
            return self._state.get_agent(self._relation_uuid)
        elif self.relation_type is RelationType.event:
            return self._state.get_event(self._relation_uuid)
        elif self.relation_type is RelationType.season:
            return self._state.get_season(self._relation_uuid)
        if self.relation_type and self._relation_uuid:
            _log.warning(f'Unknown relationship type={self.relation_type!r} uuid={self._relation_uuid!r}')
        return None

    # helper methods

    def get_all_rewards(self) -> List[Reward]:
        """:class: `List[Reward]` Returns all rewards."""
        return [level.reward for chapter in self.chapters for level in chapter.levels]


class Contract(BaseModel):
    def __init__(self, state: CacheState, data: ContractPayload) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self._data: ContractPayload = data
        self._display_name: Union[str, Dict[str, str]] = data['displayName']
        self._display_icon: Optional[str] = data['displayIcon']
        self.ship_it: bool = data['shipIt']
        self.free_reward_schedule_uuid: str = data['freeRewardScheduleUuid']
        self._content: Content = Content(self._state, data['content'])
        self.asset_path: str = data['assetPath']
        self._display_name_localized: Localization = Localization(self._display_name, locale=self._state.locale)

    def __str__(self) -> str:
        return self.display_name.locale

    def __repr__(self) -> str:
        return f'<Contract display_name={self.display_name!r}>'

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Contract) and self.uuid == other.uuid

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._display_name_localized.from_locale(locale)

    @property
    def id(self) -> UUID:
        """:class: `UUID` Returns the contract id."""
        return self.uuid

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the contract's name."""
        return self._display_name_localized

    @property
    def display_icon(self) -> Optional[Asset]:
        """:class: `Asset` Returns the contract's icon."""
        if self._display_icon is None:
            return None
        return Asset._from_url(self._state, self._display_icon)

    @property
    def content(self) -> Content:
        """:class: `Content` Returns the contract's content."""
        return self._content



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz
Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import asyncio
import logging
import sys
from typing import TYPE_CHECKING, Any, ClassVar, Coroutine, Dict, Optional, TypeVar, Union
from urllib.parse import quote as _uriquote

import aiohttp

from . import __version__, utils
from .errors import BadRequest, Forbidden, HTTPException, InternalServerError, NotFound, RateLimited

MISSING = utils.MISSING

if TYPE_CHECKING:
    from .types import (
        agents,
        buddies,
        bundles,
        ceremonies,
        competitive_tiers,
        content_tiers,
        contracts,
        currencies,
        events,
        gamemodes,
        gear,
        level_borders,
        maps,
        missions,
        player_cards,
        player_titles,
        seasons,
        sprays,
        themes,
        version,
        weapons,
    )

    T = TypeVar('T')
    Response = Coroutine[Any, Any, T]

_log = logging.getLogger(__name__)

# http-client inspired by https://github.com/Rapptz/discord.py/blob/master/discord/http.pyS


async def to_json(response: aiohttp.ClientResponse) -> Union[Dict[str, Any], str]:
    text = await response.text(encoding='utf-8')
    return utils._from_json(text)


class Route:
    BASE: ClassVar[str] = 'https://valorant-api.com/v1'

    def __init__(
        self,
        method: str,
        path: str,
        **parameters: Any,
    ) -> None:
        self.method = method
        self.path = path
        self.parameters = parameters

        url = Route.BASE + path

        if parameters:
            url = url.format_map({k: _uriquote(v) if isinstance(v, str) else v for k, v in parameters.items()})

        self.url: str = url


class HTTPClient:
    def __init__(self, session: aiohttp.ClientSession = MISSING) -> None:
        self.__session: aiohttp.ClientSession = session
        user_agent = 'valorantx (https://github.com/staciax/valorant {0}) Python/{1[0]}.{1[1]} aiohttp/{2}'
        self.user_agent: str = user_agent.format(__version__, sys.version_info, aiohttp.__version__)

    async def init(self) -> None:
        if self.__session is MISSING:
            self.__session = aiohttp.ClientSession(connector=aiohttp.TCPConnector(limit=0))

    async def request(self, route: Route, **kwargs: Any) -> Any:
        method = route.method
        url = route.url
        kwargs['headers'] = {'User-Agent': self.user_agent}

        response: Optional[aiohttp.ClientResponse] = None
        data: Optional[Union[Dict[str, Any], str]] = None

        for tries in range(5):
            try:
                async with self.__session.request(method, url, **kwargs) as response:
                    _log.debug('%s %s with %s has returned %s', method, url, kwargs.get('data'), response.status)
                    data = await to_json(response)
                    if 300 > response.status >= 200:
                        _log.debug('%s %s has received %s', method, url, data)
                        return data

                    if response.status == 400:
                        raise BadRequest(response, data)

                    # we are being rate limited
                    if response.status == 429:
                        if not response.headers.get('Via') or isinstance(data, str):
                            # Banned by Cloudflare more than likely.
                            raise HTTPException(response, data)
                        raise RateLimited(response, data)
                    elif response.status == 403:
                        raise Forbidden(response, data)
                    elif response.status == 404:
                        raise NotFound(response, data)
                    elif response.status >= 500:
                        raise InternalServerError(response, data)
                    else:
                        raise HTTPException(response, data)

            except OSError as e:
                # Connection reset by peer
                if tries < 4 and e.errno in (54, 10054):
                    await asyncio.sleep(1 + tries * 2)
                    continue
                raise

        if response is not None:
            # We've run out of retries, raise.
            if response.status >= 500:
                raise InternalServerError(response, data)

            raise HTTPException(response, data)

        raise RuntimeError('Unreachable code in HTTP handling')

    async def close(self) -> None:
        if self.__session is not MISSING:
            await self.__session.close()

    def clear(self) -> None:
        if self.__session and self.__session.closed:
            self.__session = MISSING

    async def read_from_url(self, url: str) -> bytes:
        async with self.__session.get(url) as resp:
            if resp.status == 200:
                return await resp.read()
            elif resp.status == 403:
                raise Forbidden(resp, 'cannot retrieve asset')
            elif resp.status == 404:
                raise NotFound(resp, 'asset not found')
            else:
                raise HTTPException(resp, 'failed to get asset')

    async def text_from_url(self, url: str) -> str:
        async with self.__session.get(url) as resp:
            if resp.status == 200:
                return await resp.text()
            elif resp.status == 404:
                raise NotFound(resp, 'asset not found')
            elif resp.status == 403:
                raise Forbidden(resp, 'cannot retrieve asset')
            else:
                raise HTTPException(resp, 'failed to get asset')

    # valorant-api.com

    def get_agents(self, *, language: Optional[str] = 'all', is_playable_character: bool = True) -> Response[agents.Agents]:
        params = {'isPlayableCharacter': str(is_playable_character), 'language': language}
        return self.request(Route('GET', '/agents'), params=params)

    def get_agent(
        self, uuid: str, *, language: Optional[str] = 'all', is_playable_character: bool = True
    ) -> Response[agents.AgentUUID]:
        params = {'isPlayableCharacter': str(is_playable_character), 'language': language}
        return self.request(Route('GET', '/agents/{uuid}', uuid=uuid), params=params)

    # -

    def get_buddies(self, *, language: Optional[str] = 'all') -> Response[buddies.Buddies]:
        return self.request(Route('GET', '/buddies'), params={'language': language})

    def get_buddy(self, uuid: str, *, language: Optional[str] = 'all') -> Response[buddies.BuddyUUID]:
        return self.request(Route('GET', '/buddies/{uuid}', uuid=uuid), params={'language': language})

    def get_buddy_levels(self, *, language: Optional[str] = 'all') -> Response[buddies.BuddyLevels]:
        return self.request(Route('GET', '/buddies/levels'), params={'language': language})

    def get_buddy_level(self, uuid: str, *, language: Optional[str] = 'all') -> Response[buddies.BuddyLevelUUID]:
        return self.request(Route('GET', '/buddies/levels/{uuid}', uuid=uuid), params={'language': language})

    # -

    def get_bundles(self, *, language: Optional[str] = 'all') -> Response[bundles.Bundles]:
        return self.request(Route('GET', '/bundles'), params={'language': language})

    def get_bundle(self, uuid: str, *, language: Optional[str] = 'all') -> Response[bundles.BundleUUID]:
        return self.request(Route('GET', '/bundles/{uuid}', uuid=uuid), params={'language': language})

    # -

    def get_ceremonies(self, *, language: Optional[str] = 'all') -> Response[ceremonies.Ceremonies]:
        return self.request(Route('GET', '/ceremonies'), params={'language': language})

    def get_ceremony(self, uuid: str, *, language: Optional[str] = 'all') -> Response[ceremonies.CeremonyUUID]:
        return self.request(Route('GET', '/ceremonies/{uuid}', uuid=uuid), params={'language': language})

    # -

    def get_competitive_tiers(self, *, language: Optional[str] = 'all') -> Response[competitive_tiers.CompetitiveTiers]:
        return self.request(Route('GET', '/competitivetiers'), params={'language': language})

    def get_competitive_tier(
        self, uuid: str, *, language: Optional[str] = 'all'
    ) -> Response[competitive_tiers.CompetitiveTierUUID]:
        return self.request(Route('GET', '/competitivetiers/{uuid}', uuid=uuid), params={'language': language})

    # -

    def get_content_tiers(self, *, language: Optional[str] = 'all') -> Response[content_tiers.ContentTiers]:
        return self.request(Route('GET', '/contenttiers'), params={'language': language})

    def get_content_tier(self, uuid: str, *, language: Optional[str] = 'all') -> Response[content_tiers.ContentTierUUID]:
        return self.request(Route('GET', '/contenttiers/{uuid}', uuid=uuid), params={'language': language})

    # -

    def get_contracts(self, *, language: Optional[str] = 'all') -> Response[contracts.Contracts]:
        return self.request(Route('GET', '/contracts'), params={'language': language})

    def get_contract(self, uuid: str, *, language: Optional[str] = 'all') -> Response[contracts.ContractUUID]:
        return self.request(Route('GET', '/contracts/{uuid}', uuid=uuid), params={'language': language})

    # -

    def get_currencies(self, *, language: Optional[str] = 'all') -> Response[currencies.Currencies]:
        return self.request(Route('GET', '/currencies'), params={'language': language})

    def get_currency(self, uuid: str, *, language: Optional[str] = 'all') -> Response[currencies.CurrencyUUID]:
        return self.request(Route('GET', '/currencies/{uuid}', uuid=uuid), params={'language': language})

    # -

    def get_events(self, *, language: Optional[str] = 'all') -> Response[events.Events]:
        return self.request(Route('GET', '/events'), params={'language': language})

    def get_event(self, uuid: str, *, language: Optional[str] = 'all') -> Response[events.EventUUID]:
        return self.request(Route('GET', '/events/{uuid}', uuid=uuid), params={'language': language})

    # -

    def get_game_modes(self, *, language: Optional[str] = 'all') -> Response[gamemodes.GameModes]:
        return self.request(Route('GET', '/gamemodes'), params={'language': language})

    def get_game_mode(self, uuid: str, *, language: Optional[str] = 'all') -> Response[gamemodes.GameModeUUID]:
        return self.request(Route('GET', '/gamemodes/{uuid}', uuid=uuid), params={'language': language})

    def get_game_mode_equippables(self, *, language: Optional[str] = 'all') -> Response[gamemodes.GameModeEquippables]:
        return self.request(Route('GET', '/gamemodes/equippables'), params={'language': language})

    def get_game_mode_equippable(
        self, uuid: str, *, language: Optional[str] = 'all'
    ) -> Response[gamemodes.GameModeEquippableUUID]:
        return self.request(
            Route('GET', '/gamemodes/equippables/{uuid}', uuid=uuid),
            params={'language': language},
        )

    # -

    def get_gear(self, *, language: Optional[str] = 'all') -> Response[gear.Gear]:
        return self.request(Route('GET', '/gear'), params={'language': language})

    def get_gear_(self, uuid: str, *, language: Optional[str] = 'all') -> Response[gear.GearUUID]:
        return self.request(Route('GET', '/gear/{uuid}', uuid=uuid), params={'language': language})

    # -

    def get_level_borders(self) -> Response[level_borders.LevelBorders]:
        return self.request(Route('GET', '/levelborders'))

    def get_level_border(self, uuid: str) -> Response[level_borders.LevelBorderUUID]:
        return self.request(Route('GET', '/levelborders/{uuid}', uuid=uuid))

    # -

    def get_maps(self, *, language: Optional[str] = 'all') -> Response[maps.Maps]:
        return self.request(Route('GET', '/maps'), params={'language': language})

    def get_map(self, uuid: str, *, language: Optional[str] = 'all') -> Response[maps.MapUUID]:
        return self.request(Route('GET', '/maps/{uuid}', uuid=uuid), params={'language': language})

    # -

    def get_missions(self, *, language: Optional[str] = 'all') -> Response[missions.Missions]:
        return self.request(Route('GET', '/missions'), params={'language': language})

    def get_mission(self, uuid: str, *, language: Optional[str] = 'all') -> Response[missions.MissionUUID]:
        return self.request(Route('GET', '/missions/{uuid}', uuid=uuid), params={'language': language})

    # -

    def get_player_cards(self, *, language: Optional[str] = 'all') -> Response[player_cards.PlayerCards]:
        return self.request(Route('GET', '/playercards'), params={'language': language})

    def get_player_card(self, uuid: str, *, language: Optional[str] = 'all') -> Response[player_cards.PlayerCardUUID]:
        return self.request(Route('GET', '/playercards/{uuid}', uuid=uuid), params={'language': language})

    # -

    def get_player_titles(self, *, language: Optional[str] = 'all') -> Response[player_titles.PlayerTitles]:
        return self.request(Route('GET', '/playertitles'), params={'language': language})

    def get_player_title(self, uuid: str, *, language: Optional[str] = 'all') -> Response[player_titles.PlayerTitleUUID]:
        return self.request(Route('GET', '/playertitles/{uuid}', uuid=uuid), params={'language': language})

    # -

    def get_seasons(self, *, language: Optional[str] = 'all') -> Response[seasons.Seasons]:
        return self.request(Route('GET', '/seasons'), params={'language': language})

    def get_season(self, uuid: str, *, language: Optional[str] = 'all') -> Response[seasons.SeasonUUID]:
        return self.request(Route('GET', '/seasons/{uuid}', uuid=uuid), params={'language': language})

    def get_competitive_seasons(self) -> Response[seasons.CompetitiveSeasons]:
        return self.request(Route('GET', '/seasons/competitive'))

    def get_competitive_season(self, uuid: str) -> Response[seasons.CompetitiveSeasonUUID]:
        return self.request(Route('GET', '/seasons/competitive/{uuid}', uuid=uuid))

    # -

    def get_sprays(self, *, language: Optional[str] = 'all') -> Response[sprays.Sprays]:
        return self.request(Route('GET', '/sprays'), params={'language': language})

    def get_spray(self, uuid: str, *, language: Optional[str] = 'all') -> Response[sprays.SprayUUID]:
        return self.request(Route('GET', '/sprays/{uuid}', uuid=uuid), params={'language': language})

    def get_spray_levels(self, *, language: Optional[str] = 'all') -> Response[sprays.SprayLevels]:
        return self.request(Route('GET', '/sprays/levels'), params={'language': language})

    def get_spray_level(self, uuid: str, *, language: Optional[str] = 'all') -> Response[sprays.SprayLevelUUID]:
        return self.request(Route('GET', '/sprays/levels/{uuid}', uuid=uuid), params={'language': language})

    # -

    def get_themes(self, *, language: Optional[str] = 'all') -> Response[themes.Themes]:
        return self.request(Route('GET', '/themes'), params={'language': language})

    def get_theme(self, uuid: str, *, language: Optional[str] = 'all') -> Response[themes.ThemeUUID]:
        return self.request(Route('GET', '/themes/{uuid}', uuid=uuid), params={'language': language})

    # -

    def get_weapons(self, *, language: Optional[str] = 'all') -> Response[weapons.Weapons]:
        return self.request(Route('GET', '/weapons'), params={'language': language})

    def get_weapon(self, uuid: str, *, language: Optional[str] = 'all') -> Response[weapons.WeaponUUID]:
        return self.request(Route('GET', '/weapons/{uuid}', uuid=uuid), params={'language': language})

    def get_weapon_skins(self, *, language: Optional[str] = 'all') -> Response[weapons.Skins]:
        return self.request(Route('GET', '/weapons/skins'), params={'language': language})

    def get_weapon_skin(self, uuid: str, *, language: Optional[str] = 'all') -> Response[weapons.SkinUUID]:
        return self.request(Route('GET', '/weapons/skins/{uuid}', uuid=uuid), params={'language': language})

    def get_weapon_skin_chromas(self, *, language: Optional[str] = 'all') -> Response[weapons.SkinChromas]:
        return self.request(Route('GET', '/weapons/skinchromas'), params={'language': language})

    def get_weapon_skin_chroma(self, uuid: str, *, language: Optional[str] = 'all') -> Response[weapons.SkinChromaUUID]:
        return self.request(Route('GET', '/weapons/skinchromas/{uuid}', uuid=uuid), params={'language': language})

    def get_weapon_skin_levels(self, *, language: Optional[str] = 'all') -> Response[weapons.SkinLevels]:
        return self.request(Route('GET', '/weapons/skinlevels'), params={'language': language})

    def get_weapon_skin_level(self, uuid: str, *, language: Optional[str] = 'all') -> Response[weapons.SkinLevelUUID]:
        return self.request(Route('GET', '/weapons/skinlevels/{uuid}', uuid=uuid), params={'language': language})

    # -

    def get_version(self) -> Response[version.Version]:
        return self.request(Route('GET', '/version'))



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz
Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
# - source: https://github.com/Rapptz/discord.py/blob/master/discord/utils.py

from __future__ import annotations

import datetime
import json
from typing import Any

try:
    import orjson  # type: ignore
except ImportError:
    HAS_ORJSON = False
else:
    HAS_ORJSON = True

if HAS_ORJSON:
    _from_json = orjson.loads  # type: ignore
else:
    _from_json = json.loads


class _MissingSentinel:
    __slots__ = ()

    def __eq__(self, other):
        return False

    def __bool__(self):
        return False

    def __hash__(self):
        return 0

    def __repr__(self):
        return '...'


MISSING: Any = _MissingSentinel()

# - end


def string_escape(string: str) -> str:
    # string = string.encode('raw_unicode_escape').decode('unicode_escape')
    string = string.replace('\r\n', ' ')
    string = string.replace('\t', ' ')
    string = string.replace('\r', ' ')
    string = string.replace('\n', ' ')
    string = string.replace('"', '\\"')
    return string


def parse_iso_datetime(iso: str) -> datetime.datetime:
    """Convert ISO8601 string to datetime"""
    try:
        dt = datetime.datetime.strptime(iso, '%Y-%m-%dT%H:%M:%S.%fZ')
    except ValueError:
        dt = datetime.datetime.strptime(iso, '%Y-%m-%dT%H:%M:%SZ')
    return dt.replace(tzinfo=datetime.timezone.utc)


def removeprefix(string: str, prefix: str) -> str:
    """Remove prefix from string"""
    # python 3.8 is not supported .removeprefix()
    if string.startswith(prefix):
        return string[len(prefix) :]
    return string



"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, List, Optional, Type

from .cache import CacheState
from .enums import Locale
from .http import HTTPClient
from .models.agents import Agent
from .models.buddies import Buddy, BuddyLevel
from .models.bundles import Bundle
from .models.ceremonies import Ceremony
from .models.competitive_tiers import CompetitiveTier
from .models.content_tiers import ContentTier
from .models.contracts import Contract
from .models.currencies import Currency
from .models.events import Event
from .models.gamemodes import GameMode, GameModeEquippable
from .models.gear import Gear
from .models.level_borders import LevelBorder
from .models.maps import Map
from .models.missions import Mission
from .models.player_cards import PlayerCard
from .models.player_titles import PlayerTitle
from .models.seasons import CompetitiveSeason, Season
from .models.sprays import Spray, SprayLevel
from .models.themes import Theme
from .models.version import Version
from .models.weapons import Skin, SkinChroma, SkinLevel, Weapon
from .utils import MISSING

if TYPE_CHECKING:
    from types import TracebackType

    from typing_extensions import Self

_log = logging.getLogger(__name__)

# fmt: off
__all__ = (
    'Client',
)
# fmt: on


class Client:
    def __init__(
        self,
        locale: Locale = Locale.american_english
        # *,
        # enable_cache: bool = True,
    ) -> None:
        self.locale: Locale = locale
        # self.enable_cache: bool = enable_cache
        self.http: HTTPClient = HTTPClient()
        self.cache: CacheState = CacheState(locale=locale, http=self.http)
        self._ready: asyncio.Event = MISSING
        self._closed: bool = False

    async def __aenter__(self) -> Self:
        await self.init()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        if not self.is_closed():
            await self.close()

    async def init(self) -> None:
        self._ready = asyncio.Event()
        await self.http.init()
        await self.cache.init()
        self._ready.set()
        _log.info('client initialized')

    async def reload(self) -> None:
        self.cache.clear()
        await self.cache.init()
        _log.info('client reloaded')

    def is_closed(self) -> bool:
        return self._closed

    def is_ready(self) -> bool:
        return self._ready is not MISSING and self._ready.is_set()

    async def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        if self._ready is not MISSING:
            self._ready.clear()

        self.cache.clear()
        await self.http.close()
        _log.info('client closed')

    def clear(self) -> None:
        self._closed = False
        self._ready.clear()
        self._ready = MISSING
        self.cache.clear()
        self.http.clear()

    async def wait_until_ready(self) -> None:
        if self._ready is not MISSING:
            await self._ready.wait()
        else:
            raise RuntimeError('Client not initialized yet.')

    # agents

    @property
    def agents(self) -> List[Agent]:
        return self.cache.agents

    def get_agent(self, uuid: str, /) -> Optional[Agent]:
        return self.cache.get_agent(uuid)

    async def fetch_agent(self, uuid: str, /) -> Optional[Agent]:
        await self.cache.fetch_agents()
        return self.get_agent(uuid)

    # buddies

    @property
    def buddies(self) -> List[Buddy]:
        return self.cache.buddies

    @property
    def buddy_levels(self) -> List[BuddyLevel]:
        return self.cache.buddy_levels

    def get_buddy(self, uuid: str, /) -> Optional[Buddy]:
        return self.cache.get_buddy(uuid)

    def get_buddy_level(self, uuid: str, /) -> Optional[BuddyLevel]:
        return self.cache.get_buddy_level(uuid)

    async def fetch_buddy(self, uuid: str, /) -> Optional[Buddy]:
        await self.cache.fetch_buddies()
        return self.get_buddy(uuid)

    async def fetch_buddy_level(self, uuid: str, /) -> Optional[BuddyLevel]:
        await self.cache.fetch_buddies()
        return self.get_buddy_level(uuid)

    # bundles

    @property
    def bundles(self) -> List[Bundle]:
        return self.cache.bundles

    def get_bundle(self, uuid: str, /) -> Optional[Bundle]:
        return self.cache.get_bundle(uuid)

    async def fetch_bundle(self, uuid: str, /) -> Optional[Bundle]:
        await self.cache.fetch_bundles()
        return self.get_bundle(uuid)

    # ceremonies

    @property
    def ceremonies(self) -> List[Ceremony]:
        return self.cache.ceremonies

    def get_ceremony(self, uuid: str, /) -> Optional[Ceremony]:
        return self.cache.get_ceremony(uuid)

    async def fetch_ceremony(self, uuid: str, /) -> Optional[Ceremony]:
        await self.cache.fetch_ceremonies()
        return self.get_ceremony(uuid)

    # competitive_tiers

    @property
    def competitive_tiers(self) -> List[CompetitiveTier]:
        return self.cache.competitive_tiers

    def get_competitive_tier(self, uuid: str, /) -> Optional[CompetitiveTier]:
        return self.cache.get_competitive_tier(uuid)

    async def fetch_competitive_tier(self, uuid: str, /) -> Optional[CompetitiveTier]:
        await self.cache.fetch_competitive_tiers()
        return self.get_competitive_tier(uuid)

    # content_tiers

    @property
    def content_tiers(self) -> List[ContentTier]:
        return self.cache.content_tiers

    def get_content_tier(self, uuid: str, /) -> Optional[ContentTier]:
        return self.cache.get_content_tier(uuid)

    async def fetch_content_tier(self, uuid: str, /) -> Optional[ContentTier]:
        await self.cache.fetch_content_tiers()
        return self.get_content_tier(uuid)

    # contracts

    @property
    def contracts(self) -> List[Contract]:
        return self.cache.contracts

    def get_contract(self, uuid: str, /) -> Optional[Contract]:
        return self.cache.get_contract(uuid)

    async def fetch_contract(self, uuid: str, /) -> Optional[Contract]:
        await self.cache.fetch_contracts()
        return self.get_contract(uuid)

    # currencies

    @property
    def currencies(self) -> List[Currency]:
        return self.cache.currencies

    def get_currency(self, uuid: str, /) -> Optional[Currency]:
        return self.cache.get_currency(uuid)

    async def fetch_currency(self, uuid: str, /) -> Optional[Currency]:
        await self.cache.fetch_currencies()
        return self.get_currency(uuid)

    # events

    @property
    def events(self) -> List[Event]:
        return self.cache.events

    def get_event(self, uuid: str, /) -> Optional[Event]:
        return self.cache.get_event(uuid)

    async def fetch_event(self, uuid: str, /) -> Optional[Event]:
        await self.cache.fetch_events()
        return self.get_event(uuid)

    # game_modes

    @property
    def game_modes(self) -> List[GameMode]:
        return self.cache.game_modes

    @property
    def game_mode_equippables(self) -> List[GameModeEquippable]:
        return self.cache.game_mode_equippables

    def get_game_mode(self, uuid: str, /) -> Optional[GameMode]:
        return self.cache.get_game_mode(uuid)

    def get_game_mode_equippable(self, uuid: str, /) -> Optional[GameModeEquippable]:
        return self.cache.get_game_mode_equippable(uuid)

    async def fetch_game_mode(self, uuid: str, /) -> Optional[GameMode]:
        await self.cache.fetch_game_modes()
        return self.get_game_mode(uuid)

    async def fetch_game_mode_equippable(self, uuid: str, /) -> Optional[GameModeEquippable]:
        await self.cache.fetch_game_modes()
        return self.get_game_mode_equippable(uuid)

    # gear

    @property
    def gear(self) -> List[Gear]:
        return self.cache.gear

    def get_gear(self, uuid: str, /) -> Optional[Gear]:
        return self.cache.get_gear(uuid)

    async def fetch_gear(self, uuid: str, /) -> Optional[Gear]:
        await self.cache.fetch_gear()
        return self.get_gear(uuid)

    # level_borders

    @property
    def level_borders(self) -> List[LevelBorder]:
        return self.cache.level_borders

    def get_level_border(self, uuid: str, /) -> Optional[LevelBorder]:
        return self.cache.get_level_border(uuid)

    async def fetch_level_border(self, uuid: str, /) -> Optional[LevelBorder]:
        await self.cache.fetch_level_borders()
        return self.get_level_border(uuid)

    # maps

    @property
    def maps(self) -> List[Map]:
        return self.cache.maps

    def get_map(self, uuid: str, /) -> Optional[Map]:
        return self.cache.get_map(uuid)

    async def fetch_map(self, uuid: str, /) -> Optional[Map]:
        await self.cache.fetch_maps()
        return self.get_map(uuid)

    # missions

    @property
    def missions(self) -> List[Mission]:
        return self.cache.missions

    def get_mission(self, uuid: str, /) -> Optional[Mission]:
        return self.cache.get_mission(uuid)

    async def fetch_mission(self, uuid: str, /) -> Optional[Mission]:
        await self.cache.fetch_missions()
        return self.get_mission(uuid)

    # player_cards

    @property
    def player_cards(self) -> List[PlayerCard]:
        return self.cache.player_cards

    def get_player_card(self, uuid: str) -> Optional[PlayerCard]:
        return self.cache.get_player_card(uuid)

    async def fetch_player_card(self, uuid: str) -> Optional[PlayerCard]:
        await self.cache.fetch_player_cards()
        return self.get_player_card(uuid)

    # player_titles

    @property
    def player_titles(self) -> List[PlayerTitle]:
        return self.cache.player_titles

    def get_player_title(self, uuid: str, /) -> Optional[PlayerTitle]:
        return self.cache.get_player_title(uuid)

    async def fetch_player_title(self, uuid: str, /) -> Optional[PlayerTitle]:
        await self.cache.fetch_player_titles()
        return self.get_player_title(uuid)

    # seasons

    @property
    def seasons(self) -> List[Season]:
        return self.cache.seasons

    def get_season(self, uuid: str, /) -> Optional[Season]:
        return self.cache.get_season(uuid)

    async def fetch_season(self, uuid: str, /) -> Optional[Season]:
        await self.cache.fetch_seasons()
        return self.get_season(uuid)

    @property
    def competitive_seasons(self) -> List[CompetitiveSeason]:
        return list(self.cache.competitive_seasons)

    def get_competitive_season(self, uuid: str, /) -> Optional[CompetitiveSeason]:
        return self.cache.get_competitive_season(uuid)

    def get_competitive_season_season_id(self, season_id: str, /) -> Optional[CompetitiveSeason]:
        for season in self.competitive_seasons:
            if season.season_uuid == season_id:
                return season
        return None

    async def fetch_competitive_season(self, uuid: str, /) -> Optional[CompetitiveSeason]:
        await self.cache.fetch_competitive_seasons()
        return self.get_competitive_season(uuid)

    # sprays

    @property
    def sprays(self) -> List[Spray]:
        return self.cache.sprays

    def get_spray(self, uuid: str, /) -> Optional[Spray]:
        return self.cache.get_spray(uuid)

    async def fetch_spray(self, uuid: str, /) -> Optional[Spray]:
        await self.cache.fetch_sprays()
        return self.get_spray(uuid)

    @property
    def spray_levels(self) -> List[SprayLevel]:
        return self.cache.spray_levels

    def get_spray_level(self, uuid: str, /) -> Optional[SprayLevel]:
        return self.cache.get_spray_level(uuid)

    async def fetch_spray_level(self, uuid: str, /) -> Optional[SprayLevel]:
        await self.cache.fetch_sprays()
        return self.get_spray_level(uuid)

    # themes

    @property
    def themes(self) -> List[Theme]:
        return self.cache.themes

    def get_theme(self, uuid: str) -> Optional[Theme]:
        return self.cache.get_theme(uuid)

    async def fetch_theme(self, uuid: str) -> Optional[Theme]:
        await self.cache.fetch_themes()
        return self.get_theme(uuid)

    # version

    @property
    def version(self) -> Version:
        return self.cache.version

    async def fetch_version(self) -> Version:
        await self.cache.fetch_version()
        return self.version

    # weapons

    @property
    def weapons(self) -> List[Weapon]:
        return self.cache.weapons

    def get_weapon(self, uuid: str, /) -> Optional[Weapon]:
        return self.cache.get_weapon(uuid)

    async def fetch_weapon(self, uuid: str, /) -> Optional[Weapon]:
        await self.cache.fetch_weapons()
        return self.get_weapon(uuid)

    @property
    def skins(self) -> List[Skin]:
        return self.cache.skins

    def get_skin(self, uuid: str, /) -> Optional[Skin]:
        return self.cache.get_skin(uuid)

    async def fetch_skin(self, uuid: str, /) -> Optional[Skin]:
        await self.cache.fetch_weapons()
        return self.get_skin(uuid)

    @property
    def skin_chromas(self) -> List[SkinChroma]:
        return self.cache.skin_chromas

    def get_skin_chroma(self, uuid: str, /) -> Optional[SkinChroma]:
        return self.cache.get_skin_chroma(uuid)

    async def fetch_skin_chroma(self, uuid: str, /) -> Optional[SkinChroma]:
        await self.cache.fetch_weapons()
        return self.get_skin_chroma(uuid)

    @property
    def skin_levels(self) -> List[SkinLevel]:
        return self.cache.skin_levels

    def get_skin_level(self, uuid: str, /) -> Optional[SkinLevel]:
        return self.cache.get_skin_level(uuid)

    async def fetch_skin_level(self, uuid: str, /) -> Optional[SkinLevel]:
        await self.cache.fetch_weapons()
        return self.get_skin_level(uuid)



"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional, Union

from .. import utils
from ..asset import Asset
from ..enums import MELEE_WEAPON_ID, Locale
from ..localization import Localization
from .abc import BaseModel, ShopData

if TYPE_CHECKING:
    from typing_extensions import Self

    from ..cache import CacheState
    from ..types.weapons import (
        AdsStats as AdsStatsPayload,
        AirBurstStats as AirBurstStatsPayload,
        AltShotgunStats as AltShotgunStatsPayload,
        DamageRange as DamageRangePayload,
        Skin as SkinPayload,
        SkinChroma as SkinChromaPayload,
        SkinLevel as SkinLevelPayload,
        Weapon as WeaponPayload,
        WeaponStats as WeaponStatsPayload,
    )
    from .content_tiers import ContentTier
    from .themes import Theme

__all__ = (
    'AdsStats',
    'AirBurstStats',
    'AltShotgunStats',
    'DamageRange',
    'Skin',
    'SkinChroma',
    'SkinLevel',
    'Weapon',
    'WeaponStats',
)


class AdsStats:
    def __init__(self, data: AdsStatsPayload) -> None:
        self.zoom_multiplier: float = data['zoomMultiplier']
        self.fire_rate: float = data['fireRate']
        self.run_speed_multiplier: float = data['runSpeedMultiplier']
        self.burst_count: float = data['burstCount']
        self.first_bullet_accuracy: float = data['firstBulletAccuracy']

    def __repr__(self) -> str:
        attrs = [
            ('zoom_multiplier', self.zoom_multiplier),
            ('fire_rate', self.fire_rate),
            ('run_speed_multiplier', self.run_speed_multiplier),
            ('burst_count', self.burst_count),
            ('first_bullet_accuracy', self.first_bullet_accuracy),
        ]
        joined = ' '.join('%s=%r' % t for t in attrs)
        return f'<{self.__class__.__name__} {joined}>'


class AltShotgunStats:
    def __init__(self, data: AltShotgunStatsPayload) -> None:
        self.shotgun_pellet_count: float = data['shotgunPelletCount']
        self.burst_rate: float = data['burstRate']

    def __repr__(self) -> str:
        return f"<AltShotgunStats shotgun_pellet_count={self.shotgun_pellet_count} burst_rate={self.burst_rate}>"


class AirBurstStats:
    def __init__(self, data: AirBurstStatsPayload) -> None:
        self.shotgun_pellet_count: float = data['shotgunPelletCount']
        self.burst_distance: float = data['burstDistance']

    def __repr__(self) -> str:
        return f"<AirBurstStats shotgun_pellet_count={self.shotgun_pellet_count} burst_distance={self.burst_distance}>"


class DamageRange:
    def __init__(self, data: DamageRangePayload) -> None:
        self.range_start_meters: float = data['rangeStartMeters']
        self.range_end_meters: float = data['rangeEndMeters']
        self.head_damage: float = data['headDamage']
        self.body_damage: float = data['bodyDamage']
        self.leg_damage: float = data['legDamage']

    def __repr__(self) -> str:
        attrs = [
            ('range_start_meters', self.range_start_meters),
            ('range_end_meters', self.range_end_meters),
            ('head_damage', self.head_damage),
            ('body_damage', self.body_damage),
            ('leg_damage', self.leg_damage),
        ]
        joined = ' '.join('%s=%r' % t for t in attrs)
        return f'<{self.__class__.__name__} {joined}>'


class WeaponStats:
    def __init__(self, data: WeaponStatsPayload) -> None:
        self.fire_rate: float = data['fireRate']
        self.magazine_size: int = data['magazineSize']
        self.run_speed_multiplier: float = data['runSpeedMultiplier']
        self.equip_time_seconds: float = data['equipTimeSeconds']
        self.reload_time_seconds: float = data['reloadTimeSeconds']
        self.first_bullet_accuracy: float = data['firstBulletAccuracy']
        self.shotgun_pellet_count: int = data['shotgunPelletCount']
        self._wall_penetration: Optional[str] = data['wallPenetration']
        self._feature: Optional[str] = data['feature']
        self._fire_mode: Optional[str] = data['fireMode']
        self._alt_fire_type: Optional[str] = data['altFireType']
        self.ads_stats: Optional[AdsStats] = None
        if data['adsStats'] is not None:
            self.ads_stats = AdsStats(data['adsStats'])
        self.alt_shotgun_stats: Optional[AltShotgunStats] = None
        if data['altShotgunStats'] is not None:
            self.alt_shotgun_stats = AltShotgunStats(data['altShotgunStats'])
        self.air_burst_stats: Optional[AirBurstStats] = None
        if data['airBurstStats'] is not None:
            self.air_burst_stats = AirBurstStats(data['airBurstStats'])
        self.damage_ranges: List[DamageRange] = [DamageRange(x) for x in data['damageRanges']]

    def __repr__(self) -> str:
        attrs = [
            ('fire_rate', self.fire_rate),
            ('magazine_size', self.magazine_size),
            ('run_speed_multiplier', self.run_speed_multiplier),
            ('equip_time_seconds', self.equip_time_seconds),
            ('reload_time_seconds', self.reload_time_seconds),
            ('first_bullet_accuracy', self.first_bullet_accuracy),
            ('shotgun_pellet_count', self.shotgun_pellet_count),
            ('wall_penetration', self.wall_penetration),
            ('feature', self.feature),
            ('fire_mode', self.fire_mode),
            ('alt_fire_type', self.alt_fire_type),
            ('ads_stats', self.ads_stats),
            ('alt_shotgun_stats', self.alt_shotgun_stats),
            ('air_burst_stats', self.air_burst_stats),
            ('damage_ranges', self.damage_ranges),
        ]
        joined = ' '.join('%s=%r' % t for t in attrs)
        return f'<{self.__class__.__name__} {joined}>'

    @property
    def fire_mode(self) -> Optional[str]:
        if self._fire_mode is not None:
            return utils.removeprefix(self._fire_mode, 'EWeaponFireModeDisplayType::')
        return None

    @property
    def wall_penetration(self) -> Optional[str]:
        if self._wall_penetration is not None:
            return utils.removeprefix(self._wall_penetration, 'EWallPenetrationDisplayType::')
        return None

    @property
    def feature(self) -> Optional[str]:
        if self._feature is not None:
            return utils.removeprefix(self._feature, 'WeaponStatsFeature::')
        return None

    @property
    def alt_fire_type(self) -> Optional[str]:
        if self._alt_fire_type is not None:
            return utils.removeprefix(self._alt_fire_type, 'EWeaponAltFireDisplayType::')
        return None


class Weapon(BaseModel):
    def __init__(self, *, state: CacheState, data: WeaponPayload) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self._data: WeaponPayload = data
        self._display_name: Union[str, Dict[str, str]] = data['displayName']
        self._category: str = data['category']
        self._default_skin_uuid: str = data['defaultSkinUuid']
        self._display_icon: str = data['displayIcon']
        self._kill_stream_icon: str = data['killStreamIcon']
        self.asset_path: str = data['assetPath']
        self.weapon_stats: Optional[WeaponStats] = None
        if data['weaponStats'] is not None:
            self.weapon_stats = WeaponStats(data['weaponStats'])
        self.shop_data: Optional[ShopData] = None
        if data['shopData'] is not None:
            self.shop_data = ShopData(state=self._state, item=self, data=data['shopData'])
        self.skins: List[Skin] = [Skin(state=self._state, data=skin, parent=self) for skin in data['skins']]
        self._is_melee: bool = True if self.uuid == MELEE_WEAPON_ID else False
        self._display_name_localized: Localization = Localization(self._display_name, locale=self._state.locale)

    def __str__(self) -> str:
        return self.display_name.locale

    def __repr__(self) -> str:
        return f"<Weapon display_name={self.display_name!r}>"

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._display_name_localized.from_locale(locale)

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the weapon's name."""
        return self._display_name_localized

    @property
    def category(self) -> str:
        """:class: `str` Returns the weapon's category."""
        return utils.removeprefix(self._category, "EEquippableCategory::")

    @property
    def display_icon(self) -> Asset:
        """:class: `Asset` Returns the weapon's icon."""
        return Asset._from_url(self._state, self._display_icon)

    @property
    def kill_stream_icon(self) -> Asset:
        """:class: `Asset` Returns the weapon's kill stream icon."""
        return Asset._from_url(self._state, self._kill_stream_icon)

    def is_melee(self) -> bool:
        """:class: `bool` Returns whether the weapon is a melee weapon."""
        return self._is_melee

    @property
    def stats(self) -> Optional[WeaponStats]:
        """:class: `Optional[WeaponStats]` alias for :attr: `weapon_stats`"""
        return self.weapon_stats

    @classmethod
    def _copy(cls, weapon: Self) -> Self:
        self = cls.__new__(cls)  # bypass __init__
        self._uuid = weapon._uuid
        self._state = weapon._state
        self._data = weapon._data.copy()
        self._display_name = weapon._display_name
        self._category = weapon._category
        self._default_skin_uuid = weapon._default_skin_uuid
        self._display_icon = weapon._display_icon
        self._kill_stream_icon = weapon._kill_stream_icon
        self.asset_path = weapon.asset_path
        self.weapon_stats = weapon.weapon_stats
        self.shop_data = weapon.shop_data
        self.skins = weapon.skins
        self._is_melee = weapon._is_melee
        self._display_name_localized = weapon._display_name_localized
        return self

    # helpers

    def is_random(self) -> bool:
        return 'random' in self.asset_path.lower()


class Skin(BaseModel):
    def __init__(self, *, state: CacheState, data: SkinPayload, parent: Weapon) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self._data: SkinPayload = data
        self._display_name: Union[str, Dict[str, str]] = data['displayName']
        self._theme_uuid: str = data['themeUuid']
        self._content_tier_uuid: Optional[str] = data['contentTierUuid']
        self._display_icon: str = data['displayIcon']
        self._wallpaper: Optional[str] = data['wallpaper']
        self.asset_path: str = data['assetPath']
        self.chromas: List[SkinChroma] = [
            SkinChroma(state=self._state, data=chroma, parent=self) for chroma in data['chromas']
        ]
        self.levels: List[SkinLevel] = [
            SkinLevel(state=self._state, data=level, parent=self, level_number=index)
            for index, level in enumerate(data['levels'])
        ]
        self.parent: Weapon = parent
        self._display_name_localized: Localization = Localization(self._display_name, locale=self._state.locale)

    def __str__(self) -> str:
        return self.display_name.locale

    def __repr__(self) -> str:
        return f"<Skin display_name={self.display_name!r}>"

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._display_name_localized.from_locale(locale)

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the skin's name."""
        return self._display_name_localized

    @property
    def theme(self) -> Optional[Theme]:
        """:class: `Theme` Returns the skin's theme uuid."""
        return self._state.get_theme(self._theme_uuid)

    @property
    def content_tier(self) -> Optional[ContentTier]:
        """:class: `ContentTier` Returns the skin's rarity."""
        if self._content_tier_uuid is None:
            return None
        return self._state.get_content_tier(self._content_tier_uuid)

    @property
    def rarity(self) -> Optional[ContentTier]:
        """:class: `ContentTier` alias for :attr: `content_tier`"""
        return self.content_tier

    @property
    def display_icon(self) -> Optional[Asset]:
        """:class: `Asset` Returns the skin's icon."""
        if self._display_icon is None:
            return None
        return Asset._from_url(self._state, self._display_icon)

    @property
    def display_icon_fix(self) -> Optional[Asset]:
        """:class: `Asset` Returns the skin's icon."""
        display_icon = self._display_icon or (self.levels[0].display_icon if len(self.levels) > 0 else None)
        if display_icon is None:
            return None
        return Asset._from_url(self._state, str(display_icon))

    @property
    def wallpaper(self) -> Optional[Asset]:
        """:class: `Asset` Returns the skin's wallpaper."""
        if self._wallpaper is None:
            return None
        return Asset._from_url(self._state, url=self._wallpaper)

    def is_melee(self) -> bool:
        """:class: `bool` Returns whether the bundle is a melee."""
        return self.parent.is_melee()

    def is_random(self) -> bool:
        return 'random' in self.asset_path.lower()

    @classmethod
    def _copy(cls, skin: Self) -> Self:
        self = cls.__new__(cls)  # bypass __init__
        self._uuid = skin._uuid
        self._state = skin._state
        self._data = skin._data.copy()
        self._display_name = skin._display_name
        self._theme_uuid = skin._theme_uuid
        self._content_tier_uuid = skin._content_tier_uuid
        self._display_icon = skin._display_icon
        self._wallpaper = skin._wallpaper
        self.asset_path = skin.asset_path
        self.chromas = skin.chromas.copy()
        self.levels = skin.levels.copy()
        self.parent = skin.parent
        self._display_name_localized = skin._display_name_localized
        return self

    # helpers

    def get_skin_level(self, level: int) -> Optional[SkinLevel]:
        """get the skin's level with the given level.

        Parameters
        ----------
        level: :class: `int`
            The level of the skin level to get.

        Returns
        -------
        Optional[:class: `SkinLevel`]
            The skin level with the given level.
        """
        return next((skin_level for skin_level in self.levels if skin_level.level_number == level), None)


class SkinChroma(BaseModel):
    def __init__(self, *, state: CacheState, data: SkinChromaPayload, parent: Skin) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self._data: SkinChromaPayload = data
        self._display_name: Union[str, Dict[str, str]] = data['displayName']
        self._display_icon: Optional[str] = data['displayIcon']
        self._full_render: str = data['fullRender']
        self._swatch: Optional[str] = data['swatch']
        self._streamed_video: Optional[str] = data['streamedVideo']
        self.asset_path: str = data['assetPath']
        self.parent: Skin = parent
        self._display_name_localized: Localization = Localization(self._display_name, locale=self._state.locale)

    def __str__(self) -> str:
        return self.display_name.locale

    def __repr__(self) -> str:
        return f'<SkinChroma display_name={self.display_name!r}>'

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        """Returns the skin's display name localized to the given locale."""
        return self._display_name_localized.from_locale(locale=locale)

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the skin's name."""
        return self._display_name_localized

    @property
    def display_icon(self) -> Optional[Asset]:
        """:class: `Asset` Returns the skin's icon."""
        if self._display_icon is None:
            return None
        return Asset._from_url(self._state, url=self._display_icon)

    @property
    def display_icon_fix(self) -> Optional[Asset]:
        """:class: `Asset` Returns the skin's icon with fixed white background."""

        skin = self.parent

        if skin is None:
            return self.display_icon

        display_icon = skin.display_icon
        if len(skin.levels) > 0:
            display_icon = skin.levels[0].display_icon

        weapon = skin.parent
        if weapon is None:
            return display_icon

        self_name = utils.removeprefix(self.display_name.default, 'Standard ')
        if self_name.lower() == weapon.display_name.default.lower():  # check if skin name is same as weapon name
            display_icon = weapon.display_icon or display_icon

        return display_icon

    @property
    def full_render(self) -> Optional[Asset]:
        """:class: `Asset` Returns the skin's icon full render."""
        if self._full_render is None:
            return None
        return Asset._from_url(self._state, url=self._full_render)

    @property
    def swatch(self) -> Optional[Asset]:
        """:class: `Asset` Returns the skin's swatch."""
        if self._swatch is None:
            return None
        return Asset._from_url(self._state, url=self._swatch)

    @property
    def streamed_video(self) -> Optional[Asset]:
        """:class: `Optional[Asset]` Returns the skin's video."""
        if self._streamed_video is None:
            return None
        return Asset._from_url(self._state, url=self._streamed_video)

    @property
    def video(self) -> Optional[Asset]:
        """:class: `Asset` alias for streamed_video."""
        return self.streamed_video

    @classmethod
    def _copy(cls, skin_chroma: Self) -> Self:
        """Copies the given skin_chroma with the given parent."""
        self = cls.__new__(cls)  # bypass __init__
        self._uuid = skin_chroma._uuid
        self._state = skin_chroma._state
        self._data = skin_chroma._data.copy()
        self._display_name = skin_chroma._display_name
        self._display_icon = skin_chroma._display_icon
        self._full_render = skin_chroma._full_render
        self._swatch = skin_chroma._swatch
        self._streamed_video = skin_chroma._streamed_video
        self.asset_path = skin_chroma.asset_path
        self.parent = skin_chroma.parent
        self._display_name_localized = skin_chroma._display_name_localized
        return self

    # helpers

    @property
    def theme(self) -> Optional[Theme]:
        """:class: `Theme` Returns the skin's theme uuid."""
        return self.parent.theme

    @property
    def content_tier(self) -> Optional[ContentTier]:
        """:class: `ContentTier` Returns the skin's rarity."""
        return self.parent.content_tier

    @property
    def rarity(self) -> Optional[ContentTier]:
        """:class: `ContentTier` alias for content_tier."""
        return self.content_tier

    def is_melee(self) -> bool:
        """:class: `bool` Returns whether the bundle is a melee."""
        return self.parent.is_melee()

    def is_random(self) -> bool:
        return 'random' in self.asset_path.lower()


class SkinLevel(BaseModel):
    def __init__(self, *, state: CacheState, data: SkinLevelPayload, parent: Skin, level_number: int) -> None:
        super().__init__(data['uuid'])
        self._state: CacheState = state
        self._data: SkinLevelPayload = data
        self._display_name: Union[str, Dict[str, str]] = data['displayName']
        self._level: Optional[str] = data['levelItem']
        self._display_icon: Optional[str] = data['displayIcon']
        self._streamed_video: Optional[str] = data['streamedVideo']
        self.asset_path: str = data['assetPath']
        self._level_number: int = level_number
        self._is_level_one: bool = level_number == 0
        self.parent: Skin = parent
        self._display_name_localized: Localization = Localization(self._display_name, locale=self._state.locale)

    def __str__(self) -> str:
        return str(self.display_name)

    def __repr__(self) -> str:
        return f"<SkinLevel display_name={self.display_name!r} level={self.level!r}>"

    def display_name_localized(self, locale: Optional[Union[Locale, str]] = None) -> str:
        return self._display_name_localized.from_locale(locale=locale)

    @property
    def display_name(self) -> Localization:
        """:class: `str` Returns the skin's name."""
        return self._display_name_localized

    @property
    def level_item(self) -> str:
        """:class: `str` Returns the skin's level."""
        if self._level is None:
            return 'Normal'
        return utils.removeprefix(self._level, 'EEquippableSkinLevelItem::')

    @property
    def level(self) -> str:
        """:class: `str` alias for level_item."""
        return self.level_item

    @property
    def display_icon(self) -> Optional[Asset]:
        """:class: `Asset` Returns the skin's icon."""
        if self._display_icon is None:
            return None
        return Asset._from_url(self._state, url=self._display_icon)

    @property
    def display_icon_fix(self) -> Optional[Asset]:
        """:class: `Asset` Returns the skin's icon with fixed white background."""
        display_icon = self._display_icon or self.parent.display_icon or self.parent.parent.display_icon
        if display_icon is None:
            return None
        if isinstance(display_icon, Asset):
            return display_icon
        return Asset._from_url(self._state, url=display_icon)

    @property
    def streamed_video(self) -> Optional[Asset]:
        """:class: `Asset` Returns the skin's video."""
        if self._streamed_video is None:
            return None
        return Asset._from_url(self._state, url=self._streamed_video)

    @property
    def video(self) -> Optional[Asset]:
        """:class: `Asset` alias for streamed_video."""
        return self.streamed_video

    def is_level_one(self) -> bool:
        """:class: `bool` Returns whether the skin is level one."""
        return self._is_level_one

    @classmethod
    def _copy(cls, skin_level: Self) -> Self:
        """Copies the given skin_level with the given parent."""
        self = cls.__new__(cls)  # bypass __init__
        self._uuid = skin_level._uuid
        self._state = skin_level._state
        self._data = skin_level._data.copy()
        self._display_name = skin_level._display_name
        self._level = skin_level._level
        self._display_icon = skin_level._display_icon
        self._streamed_video = skin_level._streamed_video
        self.asset_path = skin_level.asset_path
        self._level_number = skin_level._level_number
        self._is_level_one = skin_level._is_level_one
        self.parent = skin_level.parent._copy(skin_level.parent)
        self._display_name_localized = skin_level._display_name_localized
        return self

    # helpers

    @property
    def level_number(self) -> int:
        """:class: `int` Returns the skin's level number."""
        return self._level_number

    @property
    def theme(self) -> Optional[Theme]:
        """:class: `Theme` Returns the skin's theme uuid."""
        return self.parent.theme

    @property
    def content_tier(self) -> Optional[ContentTier]:
        """:class: `ContentTier` Returns the skin's rarity."""
        return self.parent.content_tier

    @property
    def rarity(self) -> Optional[ContentTier]:
        """:class: `ContentTier` alias for content_tier."""
        return self.content_tier

    def is_melee(self) -> bool:
        """:class: `bool` Returns whether the bundle is a melee."""
        return self.parent.is_melee()

    def is_random(self) -> bool:
        """:class: `bool` Returns whether the skin is random."""
        return 'random' in self.asset_path.lower()



"""
The MIT License (MIT)

Copyright (c) 2023-present STACiA

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import Dict, Optional, Union

from . import utils
from .enums import Locale

# fmt: off
__all__ = (
    'Localization',
)
# fmt: on


class Localization:
    def __init__(
        self,
        untranslated: Optional[Union[str, Dict[str, str]]],
        locale: Union[str, Locale] = Locale.american_english,
    ) -> None:
        self.untranslated = untranslated
        self._locale = locale
        if self.untranslated is None:
            self.untranslated = {}
        if isinstance(self.untranslated, str):
            self.untranslated = {str(self._locale): self.untranslated}
        default = self.untranslated.get(str(self._locale), '')
        self.ar_AE: str = self.untranslated.get('ar-AE', default)
        self.de_DE: str = self.untranslated.get('de-DE', default)
        self.en_US: str = self.untranslated.get('en-US', default)
        self.es_ES: str = self.untranslated.get('es-ES', default)
        self.es_MX: str = self.untranslated.get('es-MX', default)
        self.fr_FR: str = self.untranslated.get('fr-FR', default)
        self.id_ID: str = self.untranslated.get('id-ID', default)
        self.it_IT: str = self.untranslated.get('it-IT', default)
        self.ja_JP: str = self.untranslated.get('ja-JP', default)
        self.ko_KR: str = self.untranslated.get('ko-KR', default)
        self.pl_PL: str = self.untranslated.get('pl-PL', default)
        self.pt_BR: str = self.untranslated.get('pt-BR', default)
        self.ru_RU: str = self.untranslated.get('ru-RU', default)
        self.th_TH: str = self.untranslated.get('th-TH', default)
        self.tr_TR: str = self.untranslated.get('tr-TR', default)
        self.vi_VN: str = self.untranslated.get('vi-VN', default)
        self.zh_CN: str = self.untranslated.get('zh-CN', default)
        self.zh_TW: str = self.untranslated.get('zh-TW', default)

    def __str__(self) -> str:
        """Returns the default locale."""
        return self.locale

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Localization) and self.untranslated == other.untranslated

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.untranslated)

    def __lt__(self, other: Localization) -> bool:
        return self.locale < other.locale

    def __le__(self, other: Localization) -> bool:
        return self.locale <= other.locale

    def __gt__(self, other: Localization) -> bool:
        return self.locale > other.locale

    def __ge__(self, other: Localization) -> bool:
        return self.locale >= other.locale

    @property
    def arabic(self) -> str:
        """:class:`str`: Returns the Arabic locale."""
        return utils.string_escape(self.ar_AE)

    @property
    def german(self) -> str:
        """:class:`str`: Returns the German locale."""
        return utils.string_escape(self.de_DE)

    @property
    def english(self) -> str:
        """:class:`str`: Returns the English locale."""
        return utils.string_escape(self.en_US)

    @property
    def american_english(self) -> str:
        """:class:`str`: Returns the American English locale."""
        return utils.string_escape(self.en_US)

    @property
    def british_english(self) -> str:
        """:class:`str`: Returns the British English locale."""
        return utils.string_escape(self.en_US)

    @property
    def spanish(self) -> str:
        """:class:`str`: Returns the Spanish locale."""
        return utils.string_escape(self.es_ES)

    @property
    def spanish_mexican(self) -> str:
        """:class:`str`: Returns the Spanish Mexican locale."""
        return utils.string_escape(self.es_MX)

    @property
    def french(self) -> str:
        """:class:`str`: Returns the French locale."""
        return utils.string_escape(self.fr_FR)

    @property
    def indonesian(self) -> str:
        """:class:`str`: Returns the Indonesian locale."""
        return utils.string_escape(self.id_ID)

    @property
    def italian(self) -> str:
        """:class:`str`: Returns the Italian locale."""
        return utils.string_escape(self.it_IT)

    @property
    def japanese(self) -> str:
        """:class:`str`: Returns the Japanese locale."""
        return utils.string_escape(self.ja_JP)

    @property
    def korean(self) -> str:
        """:class:`str`: Returns the Korean locale."""
        return utils.string_escape(self.ko_KR)

    @property
    def polish(self) -> str:
        """:class:`str`: Returns the Polish locale."""
        return utils.string_escape(self.pl_PL)

    @property
    def portuguese_brazil(self) -> str:
        """:class:`str`: Returns the Portuguese Brazil locale."""
        return utils.string_escape(self.pt_BR)

    @property
    def russian(self) -> str:
        """:class:`str`: Returns the Russian locale."""
        return utils.string_escape(self.ru_RU)

    @property
    def thai(self) -> str:
        """:class:`str`: Returns the Thai locale."""
        return utils.string_escape(self.th_TH)

    @property
    def turkish(self) -> str:
        """:class:`str`: Returns the Turkish locale."""
        return utils.string_escape(self.tr_TR)

    @property
    def vietnamese(self) -> str:
        """:class:`str`: Returns the Vietnamese locale."""
        return utils.string_escape(self.vi_VN)

    @property
    def chinese_simplified(self) -> str:
        """:class:`str`: Returns the Chinese Simplified locale."""
        return utils.string_escape(self.zh_CN)

    @property
    def chinese_traditional(self) -> str:
        """:class:`str`: Returns the Chinese Traditional locale."""
        return utils.string_escape(self.zh_TW)

    @property
    def default(self) -> str:
        """:class:`str`: Returns the english locale is default."""
        if isinstance(self.untranslated, str):
            return self.untranslated
        elif isinstance(self.untranslated, dict):
            return self.untranslated.get('en-US', '')
        return ''

    @property
    def locale(self) -> str:
        """:class:`str`: Returns from your current locale."""
        if isinstance(self.untranslated, dict):
            return self.untranslated.get(str(self._locale), self.default)
        return self.default

    def from_locale(self, locale: Optional[Union[Locale, str]] = None) -> str:
        """:class:`str`: Returns the locale from the locale code."""
        if locale is None:
            return self.locale
        if isinstance(locale, Locale):
            locale = str(locale)

        if hasattr(self, locale.replace('-', '_')):
            return getattr(self, locale.replace('-', '_'))
        elif isinstance(self.untranslated, dict):
            return self.untranslated.get(locale, self.default)
        return self.default
